from typing import Any, Dict, List, Optional, Union

import antimatter.client as openapi_client
from antimatter.builders.identity_provider import PrincipalType, ProviderType
from antimatter.client import DefaultApi
from antimatter.session_mixins.serializers.identity_details import serialize_identity_principal_details_dict, \
    serialize_identity_provider_info_dict
from antimatter.session_mixins.token import exec_with_token


class OverrideDomainIdentityPrincipalDetails(openapi_client.DomainIdentityPrincipalDetails):
    """
    This override provides a local way to pass domain identity principal details
    to the openapi generated client that mitigates the pydantic serializing error
    it produces due to a bug in the generator.

    The code that the generator currently produces looks like:
    ```
    one_of_schemas: List[str] = Literal["APIKeyDomainIdentityProviderDetails", "GoogleOAuthDomainIdentityProviderDetails"]
    ```
    that will produce the error:
    ```
    /Users/daniel/.pyenv/versions/pycapsule-3.11/lib/python3.11/site-packages/pydantic/main.py:308: UserWarning: Pydantic serializer warnings:
    Expected `list[str]` but got `_LiteralGenericAlias` - serialized value may not be as expected
    return self.__pydantic_serializer__.to_python(
    ```
    """
    one_of_schemas: List[str] = [
        "DomainIdentityAPIKeyPrincipalParams",
        "DomainIdentityEmailPrincipalParams",
        "DomainIdentityHostedDomainPrincipalParams",
    ]


class IdentityProviderMixin:
    """
    Session mixin defining identity provider CRUD functionality.
    """

    def __init__(self, domain: str, client: DefaultApi, **kwargs):
        try:
            super().__init__(domain=domain, client=client, **kwargs)
        except TypeError:
            super().__init__()  # If this is last mixin, super() will be object()
        self._domain = domain
        self._client = client

    @exec_with_token
    def upsert_identity_provider(
        self,
        provider_name: str,
        provider_type: Union[str, ProviderType] = ProviderType.ApiKey,
        client_id: Optional[str] = None,
    ) -> Dict[str, Any]:
        """
        Create or update an identity provider.

        :param provider_name: The name of a new or existing identity provider
        :param provider_type: The provider type for identity management
        :param client_id: If the provider type is 'GoogleOAuth', a client ID must be provided
        :return: The identity provider summary
        """
        provider_type = ProviderType(provider_type)
        provider_params = None
        match ProviderType(provider_type):
            case ProviderType.GoogleOAuth:
                provider_params = openapi_client.GoogleOAuthDomainIdentityProviderDetails(
                    type=ProviderType.GoogleOAuth.value,
                    client_id=client_id,
                )
            case ProviderType.ApiKey:
                provider_params = openapi_client.APIKeyDomainIdentityProviderDetails(
                    type=ProviderType.ApiKey.value,
                )

        # Disable warnings because of a bug in DomainIdentityProviderDetails autogenerated code
        res = self._client.domain_upsert_identity_provider(
            domain_id=self._domain,
            identity_provider_name=provider_name,
            domain_identity_provider_details=openapi_client.DomainIdentityProviderDetails(provider_params)
        ).model_dump(warnings=False)
        return serialize_identity_provider_info_dict(res)

    @exec_with_token
    def insert_identity_provider_principal(
        self,
        provider_name: str,
        capabilities: Dict[str, str],
        principal_type: Union[str, PrincipalType],
        principal_value: Optional[str] = None,
    ) -> Dict[str, str]:
        """
        Creates a new principal for the provider. Note that the provider_name
        must refer to an existing identity provider. The principal_value is
        optional if the type is APIKey.

        :param provider_name: The name of an existing identity provider
        :param capabilities: The capabilities to attach to the principal
        :param principal_type: The type of principal to create. One of 'APIKey', 'Email', or 'HostedDomain'
        :param principal_value: The appropriate identifying value for the principal, depending on type
        :return: The ID of the inserted principal and any additional metadata
        """
        principal_type = PrincipalType(principal_type)
        inner_params = None
        match PrincipalType(principal_type):
            case PrincipalType.ApiKey:
                inner_params = openapi_client.DomainIdentityAPIKeyPrincipalParams(
                    type=principal_type.value, api_key_id=principal_value,
                )
            case PrincipalType.Email:
                inner_params = openapi_client.DomainIdentityEmailPrincipalParams(
                    type=principal_type.value, email=principal_value,
                )
            case PrincipalType.HostedDomain:
                inner_params = openapi_client.DomainIdentityHostedDomainPrincipalParams(
                    type=principal_type.value, hosted_domain=principal_value,
                )
        return self._client.domain_insert_identity_provider_principal(
            domain_id=self._domain,
            identity_provider_name=provider_name,
            domain_identity_provider_principal_params=openapi_client.DomainIdentityProviderPrincipalParams(
                capabilities=[openapi_client.Capability(name=k, value=v) for k, v in capabilities.items()],
                details=OverrideDomainIdentityPrincipalDetails(inner_params)
            ),
        ).model_dump(exclude_none=True)

    @exec_with_token
    def update_identity_provider_principal(
        self,
        provider_name: str,
        principal_id: str,
        capabilities: Dict[str, str],
    ) -> None:
        """
        Update the capabilities for an identity provider principal.

        :param provider_name: The name of an existing identity provider
        :param principal_id: The ID of the principal
        :param capabilities: The capabilities to attach to the principal
        """
        self._client.domain_update_identity_provider_principal(
            domain_id=self._domain,
            identity_provider_name=provider_name,
            principal_id=principal_id,
            capability_list=openapi_client.CapabilityList(
                capabilities=[openapi_client.Capability(name=k, value=v) for k, v in capabilities.items()]
            ),
        )

    @exec_with_token
    def get_identity_provider(self, provider_name: str) -> Dict[str, Any]:
        """
        Retrieve detailed information and configuration of an identity provider

        :param provider_name: The name of an existing identity provider
        :return: The identity provider details
        """
        res = self._client.domain_get_identity_provider(
            domain_id=self._domain, identity_provider_name=provider_name,
        ).model_dump(warnings=False)
        return serialize_identity_provider_info_dict(res)

    @exec_with_token
    def list_identity_providers(self) -> List[Dict[str, Any]]:
        """
        Retrieve the domain's identity providers and a brief overview of their
        configuration.
        """
        res = self._client.domain_list_identity_providers(domain_id=self._domain)
        if not res.identity_providers:
            return []

        # Disable warnings because of a bug in DomainIdentityProviderDetails autogenerated code
        return [
            serialize_identity_provider_info_dict(prov.model_dump(warnings=False))
            for prov in res.identity_providers
        ]

    @exec_with_token
    def get_identity_provider_principal(
        self,
        provider_name: str,
        principal_id: Optional[str] = None,
    ) -> Union[List[Dict[str, Any]], Dict[str, Any]]:
        """
        Get either a summary of all the principals for an identity provider, or
        detailed information about a single principal if a principal_id is
        provided

        :param provider_name: The name of an existing identity provider
        :param principal_id: The ID of the principal; None to get all principals
        :return: The principal information
        """
        if principal_id is None:
            res = self._client.domain_get_identity_provider_principals(
                domain_id=self._domain, identity_provider_name=provider_name,
            )
            principals = []
            for principal in res.principals:
                principal = principal.model_dump()
                if t := principal.get("principal_type"):
                    principal["principal_type"] = t.value
                principals.append(principal)
            return principals
        else:
            res = self._client.domain_get_identity_provider_principal(
                domain_id=self._domain,
                identity_provider_name=provider_name,
                principal_id=principal_id,
            )
            return serialize_identity_principal_details_dict(res.model_dump(warnings=False))

    @exec_with_token
    def delete_identity_provider(self, provider_name: str) -> None:
        """
        Delete an identity provider. All domain tokens created using this
        identity provider will be invalidated. Take care not to remove the
        identity provider that is providing you admin access to your domain, as
        you may lock yourself out.

        :param provider_name: The name of the identity provider to fully delete
        """
        self._client.domain_delete_identity_provider(domain_id=self._domain, identity_provider_name=provider_name)

    @exec_with_token
    def delete_identity_provider_principal(
        self, provider_name: str, principal_id: str,
    ) -> None:
        """
        Delete an identity provider principal.

        :param provider_name: The name of the identity provider to delete a principal from
        :param principal_id: The ID of the principal to delete
        """
        self._client.domain_delete_identity_provider_principal(
            domain_id=self._domain,
            identity_provider_name=provider_name,
            principal_id=principal_id,
        )
